/*
Module : PushPin.cpp
Purpose: Implementation of a push pin MFC button class
Created: PJN / 04-04-1996
History: PJN / 08-06-1996 Removed win32sup.h include
         PJN / 20-11-1997 Major Update to code, changes include
                          1. Support non default color schemes
                          2. Supports dynamic changes to color schemes
                          3. Complete restructuring of code
                          4. Can now use an edged bitmap if you want to
                          5. Dropped support for Win16
                          6. Redid the example program 
                          7. Simplified external usage of class
         PJN / 24-11-1997 1. Minor changes to support CPushPinFrame class
         PJN / 07-12-1997 Minor changes to fix a small redraw bug
         PJN / 22-02-1999 1. A number of small code updates
         PJN / 02-06-1999 1. Added a number of comments to improve code
                          readability.
                          2. Changed the resource id of the button bitmaps to more
                          meaningful values.
                          3. Class no longer uses message reflection to
                          catch changes in the clicked state. This means that
                          parent classes can now catch the clicked notification.
                          4. Code now uses mouse capturing which means that the
                          button appears pressed if you click and hold the left
                          mouse button over the push pin. This gives the pushpin
                          class this standard button behaviour.
         PJN / 14-05-2000 Button now sends WM_COMMAND messages to the parent as a
                          standard button does.
         PJN / 07-03-2001 1. Updated copyright information   
                          2. Updated the pushpin frame and pushpin classes to behave exactly
                          like their counterparts in MS Developer Studio. Thanks to Abdulah DEDIC
                          for doing all the hard work.
         PJN / 27-08-2001 1. Fixed Access violation which was occurring in PreSubclassWindow on 
                          Windows 98.
         PJN / 27-12-2016 1. Updated copyright details.
                          2. Updated the sample apps main icon
                          3. Replaced all occurrences of NULL with nullptr throughout the codebase
                          3. Updated the code to clean compile on VC 2010 to VC 2015. The code now
                          requires VC 2010 at a minimum.
         PJN / 31-12-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes
                          mean that the code will now only compile on VC 2017 or later.
         PJN / 05-06-2019 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2019
         PJN / 18-11-2019 1. Updated initialization of various structs to use C++ 11 list initialization
         PJN / 30-12-2019 1. Fixed various Clang-Tidy static code analysis warnings in the code.
         PJN / 28-03-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 03-05-2020 1. Fixed more Clang-Tidy static code analysis warnings in the code.

Copyright (c) 1997 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "resource.h"
#include "PushPin.h"


//////////////////// Macros ///////////////////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif


//////////////////// Implementation ///////////////////////////////////////////

#pragma warning(suppress: 26433 26440)
BEGIN_MESSAGE_MAP(CPushPinButton, CButton) //NOLINT(modernize-avoid-c-arrays)
  ON_WM_LBUTTONDOWN()
  ON_WM_LBUTTONUP()
  ON_WM_MOUSEMOVE()
  ON_WM_CAPTURECHANGED()
  ON_MESSAGE(WM_MOUSELEAVE, &CPushPinButton::OnMouseLeave)
END_MESSAGE_MAP()

CPushPinButton::CPushPinButton() noexcept : m_MaxRect{0, 0, 0, 0},
                                            m_bCaptured{false},
                                            m_buttonState{ButtonState::UNPINNED_NORMAL},
                                            m_bTrackLeave{false},
                                            m_uiCombnBmpID{0},
                                            m_nButtons{0}
{
}

bool CPushPinButton::ReloadBitmaps()
{
  //Reload the bitmaps
  if (!LoadBitmaps())
    return false;

  //Size to content
  if (!SizeToContent())
    return false;

  //Invalidate the maximum rect of the pushpin on the parent window
  CWnd* pParent{GetParent()};
#pragma warning(suppress: 26496)
  AFXASSUME(pParent != nullptr);
  pParent->InvalidateRect(m_MaxRect);

  //Force this button to redraw as well
  Invalidate(TRUE);

  return true;
}

BOOL CPushPinButton::SetBitmapIDs(_In_ UINT uiCombnBmpID, _In_ bool b6Buttons)
{
  m_nButtons = b6Buttons ? static_cast<int>(ButtonState::MAX_BUTTON_INDEX) : static_cast<int>(ButtonState::MAX_BUTTON_INDEX)/2;
  m_uiCombnBmpID = uiCombnBmpID;
  return LoadBitmaps();
}

BOOL CPushPinButton::Create(_In_ const RECT& rect, _In_opt_ CWnd* pParentWnd, _In_ UINT nID)
{
  //Let the base class do its thing
  if (!__super::Create(_T(""), BS_OWNERDRAW | WS_CHILD | WS_VISIBLE, rect, pParentWnd, nID))
    return FALSE;

  return ReloadBitmaps();
}

bool CPushPinButton::LoadBitmaps()
{
  //free the bitmap resources if required
  if (m_bmpCombined.operator HANDLE())
    m_bmpCombined.DeleteObject();

  //Load up the necessary bitmaps using the SDK call LoadImage
  if (!m_bmpCombined.LoadMappedBitmap(m_uiCombnBmpID))
    return false;
#ifdef _DEBUG
  BITMAP bitMap{};
  if (m_bmpCombined.GetBitmap(&bitMap))
  {
    ASSERT((bitMap.bmWidth % m_nButtons) == 0); //You're bitmap is not of the correct width
  }
#endif //#ifdef _DEBUG

  return true;
}

void CPushPinButton::DrawItem(LPDRAWITEMSTRUCT lpDIS)
{
  //Validate our parameters
#pragma warning(suppress: 26496)
  AFXASSUME(lpDIS != nullptr);

  //draw the bitmap into a compatible memory DC
#pragma warning(suppress: 26486)
  CDC* pDC{CDC::FromHandle(lpDIS->hDC)};
  if (pDC == nullptr)
    return;
  CDC memDC;
#pragma warning(suppress: 26486)
  if (!memDC.CreateCompatibleDC(pDC))
    return;
  CBitmap* pOld{memDC.SelectObject(&m_bmpCombined)};

  //Then bitblt it onto the screen DC
  CRect rect;
  rect.CopyRect(&lpDIS->rcItem);
  BITMAP bitMap{};
  if (m_bmpCombined.GetBitmap(&bitMap))
  {
    const int nButtonWidth{bitMap.bmWidth / m_nButtons};
#pragma warning(suppress: 26486)
    pDC->BitBlt(rect.left, rect.top, nButtonWidth, bitMap.bmHeight, &memDC, nButtonWidth*static_cast<int>(m_buttonState), 0, SRCCOPY);
  }
  memDC.SelectObject(pOld);
}

bool CPushPinButton::IsPinned() const noexcept
{
  return (m_buttonState == ButtonState::PINNED_NORMAL) || (m_buttonState == ButtonState::PINNED_DOWN) || (m_buttonState == ButtonState::PINNED_FLYBY);
}

void CPushPinButton::SetPinned(_In_ bool bPinned)
{
  if (m_nButtons < static_cast<int>(ButtonState::MAX_BUTTON_INDEX)) //This button does not support pinning as
    return;                                                         //if does not have the correct number of bitmaps

  if (bPinned)
    m_buttonState = ButtonState::PINNED_NORMAL;
  else
    m_buttonState = ButtonState::UNPINNED_NORMAL;

  Invalidate(TRUE);
}

void CPushPinButton::PreSubclassWindow()
{
  //Let the parent class do its thing
  __super::PreSubclassWindow();

  //Pushpin buttons must be owner draw
  ASSERT(GetWindowLong(m_hWnd, GWL_STYLE) & BS_OWNERDRAW);

  SizeToContent();
}

bool CPushPinButton::SizeToContent()
{
  //Validate our parameters
  ASSERT(m_bmpCombined.m_hObject != nullptr);

  //Get the size of the bitmap
  BITMAP bmInfo{};
  if (m_bmpCombined.GetObject(sizeof(bmInfo), &bmInfo) == 0)
    return false;
  CSize szBmp{bmInfo.bmWidth,bmInfo.bmHeight};
  if (m_uiCombnBmpID)
    szBmp.cx /= m_nButtons;

  //Calculate the client rect in parent coordinates of the maximum size of the pushpin button
  m_MaxRect = CRect{0, 0, max(szBmp.cx, m_MaxRect.Width()), max(szBmp.cy, m_MaxRect.Height())};
  ClientToScreen(&m_MaxRect);
  const CWnd* pParent{GetParent()};
#pragma warning(suppress: 26496)
  AFXASSUME(pParent != nullptr);
  CPoint p1{m_MaxRect.left, m_MaxRect.top};
  pParent->ScreenToClient(&p1);
  CPoint p2{m_MaxRect.right, m_MaxRect.bottom};
  pParent->ScreenToClient(&p2);
  m_MaxRect = CRect(p1, p2);

  //resize the button to match the size of the bitmap
  SetWindowPos(nullptr, -1, -1, szBmp.cx, szBmp.cy, SWP_NOMOVE | SWP_NOZORDER | SWP_NOREDRAW | SWP_NOACTIVATE);

  return true;
}

#pragma warning(suppress: 26434)
void CPushPinButton::OnLButtonDown(UINT nFlags, CPoint point)
{
  //Capture the mouse
  SetCapture();
  m_bCaptured = true;

  //Change the button state
  if (!IsPinned())
    m_buttonState = ButtonState::UNPINNED_DOWN;
  Invalidate();

  //Let the parent class do its thing
  __super::OnLButtonDown(nFlags, point);
}

#pragma warning(suppress: 26434)
void CPushPinButton::OnLButtonUp(UINT nFlags, CPoint point)
{
  //Release the capture
  ReleaseCapture();
  m_bCaptured = false;

  //Toggle the state of the pin and force a redraw if over the button
  CRect r;
  GetClientRect(&r);
  const BOOL bInRect{r.PtInRect(point)};
  if (bInRect)
  {
    //Update the UI
    SetPinned(!IsPinned());

    //Inform the notification window by sending the WM_COMMAND message to the parents manually
    const CWnd* pParent{GetParent()};
#pragma warning(suppress: 26496)
    AFXASSUME(pParent != nullptr);
    HWND hWnd{GetSafeHwnd()};
#pragma warning(suppress: 26490)
    pParent->SendMessage(WM_COMMAND, MAKEWPARAM(GetWindowLong(hWnd, GWL_ID), 0), reinterpret_cast<LPARAM>(hWnd));
  }
  else
  {
    //Update the UI
    SetPinned(IsPinned());
  }

  //Let the parent class do its thing
  __super::OnLButtonUp(nFlags, point);
}

#pragma warning(suppress: 26434)
void CPushPinButton::OnMouseMove(UINT nFlags, CPoint point)
{
  CRect r;
  GetClientRect(&r);
  if (r.PtInRect(point))
  {
    if (IsPinned())
    {
      if (m_buttonState != ButtonState::PINNED_DOWN)
      {
        m_buttonState = ButtonState::PINNED_DOWN;
        Invalidate(FALSE);
      }
    }
    else
    {
      if (!m_bCaptured)
      {
        if (m_buttonState != ButtonState::UNPINNED_FLYBY)
        {
          m_buttonState = ButtonState::UNPINNED_FLYBY;
          Invalidate(FALSE);
        }
      }
      else
      {
        if (m_buttonState != ButtonState::UNPINNED_DOWN)
        {
          m_buttonState = ButtonState::UNPINNED_DOWN;
          Invalidate(FALSE);
        }
      }
    }
  }
  else
  {
    if (!IsPinned())
    {
      if (m_buttonState != ButtonState::UNPINNED_FLYBY)
      {
        m_buttonState = ButtonState::UNPINNED_FLYBY;
        Invalidate();
      }
    }
  }

  //Ask that we be informed when the mouse leaves the button
  if (!m_bTrackLeave)
  {
    //First time mouse entered my window
    //request leave notification
    TRACKMOUSEEVENT tme{};
    tme.cbSize = sizeof(tme);
    tme.hwndTrack = m_hWnd;
    tme.dwFlags = TME_LEAVE;
    _TrackMouseEvent(&tme);
    m_bTrackLeave = true;
  }

  //Let the parent class do its thing
  __super::OnMouseMove(nFlags, point);
}

#pragma warning(suppress: 26434)
void CPushPinButton::OnCaptureChanged(CWnd* pWnd)
{
  //Release the capture
  if (pWnd != this)
  {
    ReleaseCapture();
    m_bCaptured = false;
    Invalidate();
  }

  //Let the parent class do its thing
  __super::OnCaptureChanged(pWnd);
}

#pragma warning(suppress: 26434)
LPARAM CPushPinButton::OnMouseLeave(WPARAM /*wParam*/, LPARAM /*lParam*/)
{
  //Return the button to its non flyby state
  if (IsPinned())
    m_buttonState = ButtonState::PINNED_NORMAL;
  else
    m_buttonState = ButtonState::UNPINNED_NORMAL;
  m_bTrackLeave = false;
  Invalidate();

  return 0L;
}
